#![no_std]
#![no_main]

use aya_bpf::{
    bindings::xdp_action,
    macros::{map, xdp},
    maps::HashMap,
    programs::XdpContext,
};
use aya_log_ebpf::info;

use responder_ebpf::util::*;
use responder_ebpf::bindings::icmphdr;
use core::mem;

const IPPROTO_ICMP: u8 = 0x01;
const ICMP_ECHO: u8 = 0x08;
const ICMP_ECHOREPLY: u8 = 0x00;
const ICMP_HDR_LEN: usize = mem::size_of::<icmphdr>();


#[inline(always)]
fn parse_icmphdr(ctx: &XdpContext, cursor: &mut usize) -> Option<*mut icmphdr> {
    let icmp = ptr_at_mut::<icmphdr>(&ctx, *cursor);
    if icmp.is_some() {
        *cursor += ICMP_HDR_LEN;
    }
    icmp
}


#[xdp(name="responder")]
pub fn responder(ctx: XdpContext) -> u32 {
    match try_responder(ctx) {
        Ok(ret) => ret,
        Err(_) => xdp_action::XDP_ABORTED,
    }
}

#[inline(always)]
unsafe fn bounce_icmp(_ctx: &XdpContext, icmp: *mut icmphdr) {
    (*icmp).type_ = ICMP_ECHOREPLY;
    (*icmp).checksum += (ICMP_ECHO - ICMP_ECHOREPLY) as u16;
}

#[map(name = "FILTER_MAP")]
static FILTER_MAP: HashMap<u32, u8> =
    HashMap::<u32, u8>::with_max_entries(0x4000000, 0);

#[inline(always)]
unsafe fn matches_filter(daddr: IpAddr) -> bool {
    match daddr {
        IpAddr::V4(daddr) => {
            if let Some(b) = FILTER_MAP.get(&daddr) {
                *b == 1u8
            } else {
                false
            }
        }
        IpAddr::V6(_daddr) => {
            false // TODO
        }
    }
}

fn try_responder(ctx: XdpContext) -> Result<xdp_action::Type, xdp_action::Type> {
    let mut hdr_cursor = 0usize;

    let (eth, ip) = unsafe {
        parse_routing(&ctx, &mut hdr_cursor)
            .ok_or(xdp_action::XDP_PASS)?
    };

    let protocol = unsafe { l3_get_protocol(&ip) };
    let daddr = unsafe { l3_get_daddr(&ip) };

    if is_local(daddr) {
        return Ok(xdp_action::XDP_PASS); // Pass normal loopback traffic to not disturb host
    }

    if unsafe { !matches_filter(daddr) } {
        info!(&ctx, "no match: drop");
        return Ok(xdp_action::XDP_DROP);
    }

    if protocol != IPPROTO_ICMP {
        return Ok(xdp_action::XDP_PASS);
    }

    let icmp = parse_icmphdr(&ctx, &mut hdr_cursor).ok_or(xdp_action::XDP_PASS)?;
    let icmp_type = unsafe { (*icmp).type_ };

    match daddr {
        IpAddr::V4(ip) => info!(&ctx, "Received matching packet with daddr: {:ipv4}", ip),
        IpAddr::V6(ip) => unsafe { info!(&ctx, "Received matching packet with daddr: {:ipv6}", ip.in6_u.u6_addr8) }
    }
    info!(&ctx, "and icmp with type: {}", icmp_type);

    if icmp_type != ICMP_ECHO {
        return Ok(xdp_action::XDP_PASS);
    }

    unsafe {
        bounce_routing(&ctx,eth, ip);
        bounce_icmp(&ctx, icmp);
    }

    Ok(xdp_action::XDP_PASS)
}

#[panic_handler]
fn panic(_info: &core::panic::PanicInfo) -> ! {
    unsafe { core::hint::unreachable_unchecked() }
}
