use crate::bindings::{ethhdr, iphdr, ipv6hdr, in6_addr};
use aya_bpf:: programs::XdpContext;

use core::mem;

pub const ETH_P_IP: u16 = 0x0800;
pub const ETH_P_IPV6: u16 = 0x86DD;
pub const ETH_HDR_LEN: usize = mem::size_of::<ethhdr>();
pub const IP_HDR_LEN: usize = mem::size_of::<iphdr>();
pub const IPV6_HDR_LEN: usize = mem::size_of::<ipv6hdr>();
pub const ETH_ALEN: usize = 6;
pub const SPOOF_SOURCE_MAC: [u8; 6] = [0xFE, 0x00, 0x00, 0x00, 0x00, 0x00];

#[inline(always)]
pub fn ptr_at<T>(ctx: &XdpContext, offset: usize) -> Option<*const T> {
    let start = ctx.data();
    let end = ctx.data_end();
    let len = mem::size_of::<T>();

    if start + offset + len > end {
        return None;
    }

    Some((start + offset) as *const T)
}

#[inline(always)]
pub fn ptr_at_mut<T>(ctx: &XdpContext, offset: usize) -> Option<*mut T> {
    let p = ptr_at::<T>(ctx, offset)?;
    Some(p as *mut T)
}

pub enum Layer3 {
    Ipv4(*mut iphdr),
    Ipv6(*mut ipv6hdr),
}

#[inline(always)]
pub fn parse_ethhdr(ctx: &XdpContext, cursor: &mut usize) -> Option<*mut ethhdr> {
    let eth = ptr_at_mut::<ethhdr>(&ctx, *cursor);
    if eth.is_some() {
        *cursor += ETH_HDR_LEN;
    }
    eth
}

#[inline(always)]
pub fn parse_ipv4hdr(ctx: &XdpContext, cursor: &mut usize) -> Option<*mut iphdr> {
    let ip = ptr_at_mut::<iphdr>(&ctx, *cursor);
    if ip.is_some() {
        *cursor += IP_HDR_LEN;
    }
    ip
}

#[inline(always)]
pub fn parse_ipv6hdr(ctx: &XdpContext, cursor: &mut usize) -> Option<*mut ipv6hdr> {
    let ip = ptr_at_mut::<ipv6hdr>(&ctx, *cursor);
    if ip.is_some() {
        *cursor += IPV6_HDR_LEN;
    }
    ip
}

#[inline(always)]
pub unsafe fn parse_routing(ctx: &XdpContext, cursor: &mut usize) -> Option<(*mut ethhdr, Layer3)> {
    if let Some(eth) = parse_ethhdr(ctx, cursor) {
        let protocol = unsafe { u16::from_be((*eth).h_proto)};
        return match protocol {
            ETH_P_IP => parse_ipv4hdr(ctx, cursor).map(|ip| (eth, Layer3::Ipv4(ip))),
            ETH_P_IPV6 => parse_ipv6hdr(ctx, cursor).map(|ip| (eth, Layer3::Ipv6(ip))),
            _ => None
        }
    }
    return None;
}

#[inline(always)]
pub unsafe fn l3_get_protocol(l3: &Layer3) -> u8 {
    return match l3 {
        Layer3::Ipv4(ip) => (**ip).protocol,
        Layer3::Ipv6(ip) => (**ip).nexthdr,
    }
}

#[inline(always)]
pub unsafe fn l3_get_daddr(l3: &Layer3) -> IpAddr {
    return match l3 {
        Layer3::Ipv4(ip) => IpAddr::V4(u32::from_be((**ip).daddr)),
        Layer3::Ipv6(ip) => IpAddr::V6((**ip).daddr),
    }
}

#[inline(always)]
pub unsafe fn bounce_eth(_ctx: &XdpContext, eth: *mut ethhdr) {
    (*eth).h_dest = (*eth).h_source;
    (*eth).h_source = SPOOF_SOURCE_MAC;
}

#[inline(always)]
pub unsafe fn bounce_ip(_ctx: &XdpContext, ip: *mut iphdr) {
    mem::swap(&mut (*ip).daddr, &mut (*ip).saddr);
    // checksum recomputation not neccessary because of swap
}

#[inline(always)]
pub unsafe fn bounce_ipv6(_ctx: &XdpContext, ip: *mut ipv6hdr) {
    mem::swap(&mut (*ip).daddr, &mut (*ip).saddr);

}

#[inline(always)]
pub unsafe fn bounce_routing(_ctx: &XdpContext, eth: *mut ethhdr, ip: Layer3) {
    bounce_eth(_ctx, eth);
    match ip {
        Layer3::Ipv4(ip) => bounce_ip(_ctx, ip),
        Layer3::Ipv6(ip) => bounce_ipv6(_ctx, ip)
    }
}

#[derive(Clone, Copy)]
pub enum IpAddr {
    V4(u32), V6(in6_addr)
}

#[inline(always)]
pub fn is_local(daddr: IpAddr) -> bool {
    match daddr {
        IpAddr::V4(daddr) => (daddr & 0xFF000000) == 0x7F000000,
        IpAddr::V6(_) => false,
    }
    // 127.0.0.0/8 addresses pass
    // ipv6 has no addresses on loopback
}

#[inline(always)]
pub fn ones_complement_add_u16(a: u16, b: u16) -> u16 {
    let (c, o) = a.overflowing_add(b);
    c + (o as u16)
}
